/*****************************************************************************/
/*
                                  HTTPd.c

WASD VMS Hypertext Transfer Protocol daemon.


COPYRIGHT NOTICE
----------------
WASD VMS Hypertext Services, Copyright (C) 1996-2001 Mark G.Daniel.
(prior to 01-JUL-1997, "HFRD VMS Hypertext Services")

This package is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; version 2 of the License, or any later version. 

This package is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
details. 

You should have received a copy of the GNU General Public License along with 
this package; if not, write to the Free Software Foundation, Inc., 675 Mass 
Ave, Cambridge, MA 02139, USA. 


PRIVILEGES REQUIRED
-------------------
These should be INSTALLed against the image.

ALTPRI   Allows the server account to raise it's prioity above 4 if enabled by
         the /PRIORITY= qualifier.

DETACH   (IMPERSONATE) Allows the server to impersonate specific accounts
         for scripting and startup purposes.

PRMGBL   Create the permanent global section used by the HTTPDMON utility.

PRMMBX   Used by the DCL scripting module to create permanent mailboxes

PSWAPM   Allows the server process to prevent itself from being swapped out if 
         enabled by the /[NO]SWAP qualifier.

SHMEM    Allow shared section memory between multiple processors
         (VAX only ,for access to the global section used by HTTPDMON).

SYSGBL   Create the system global section used by the HTTPDMON utility.

SYSLCK   Allows SYSTEM locks to be enqueued for sending commands to all
         server processes on a node/cluster (via /DO=/ALL or Admin Menu).

SYSPRV   Used for various purposes, including creating sockets within the
         privileged port range (1-1023, which includes port 80 of course),
         accessing configuration files (which can be protected from world
         access), enable AUTHORIZED write access to the file system, checking
         authentication details, amongst others.

SYSNAM   Allows the server to write into the LNM$SYSTEM logical name table.

WORLD    (optional) Allows the CONTROL module to obtain information about
         the process sending it commands.


QUALIFIERS  (implemented in the CLI.c module)
----------
/ACCEPT=                comma separated list of accepted hosts/domains
/ALL[=resname]          /do= this to all server processes on node/cluster
/AUTHORIZE=[SSL|ALL]    authorization may only be used with "https:",
                        all paths must be authorized
/CGI_PREFIX=            prefix to CGI-script variable (symbol) names
/DBUG                   turn on all "if (Debug)" statements (if compiled DBUG)
/DETACH=                DCL procedure to be run as /USER=
/DO=                    command to be passed to server
/FILBUF=                number of bytes in record buffer
/FORMAT=                log record format
/GBLSEC=DELETE=port     delete the specified global section
/IDENT=                 VMS rights identifier name used for detached processes
/[NO]LOG[=]             logging enabled/disabled, optional log file name
/[NO]MONITOR            monitor enabled/disabled
/NETBUF=                number of bytes in network read buffer
/[NO]ODS5               explicit extended file specification (testing only)
/OUTBUF=                number of bytes in output buffer
/PERIOD=                log file period
/PERSONA[=ident]        enable PERSONA services for DCL scripting,
                        optiona identifier name controlling persona access
/PORT=                  server IP port number (overriden by [service] config)
/PRIORITY=              process priority
/[NO]PROFILE            allow/disallow SYSUAF-authenticated access control
/[NO]PROMISCUOUS[=pwd]  authenticates any user name for path authorization
                        (optional password required for authorization) 
/REJECT=                comma separated list of rejected hosts/domains
/SERVICES=              list of [host-name:]port providing HTTP service
/SOFTWAREID=            overrides the server's software ID string
/SYSPRV                 normally operate with SYSPRV enabled (CAUTION!)
/[NO]SSL=               enables Secure Sockets Layer, sets protocol parameters
/[NO]SYSUAF[=ID,PROXY,RELAXED,SSL]
                        allow/disallow SYSUAF authentication, or
                        SYSUAF authentication is allowed by identifier,
                        proxy SYSUAF authentication is allowed,
                        SYSUAF authentication allowed with any current account,
                        SYSUAF authentication is only allowed with "https:"
/[NO]SWAP               allow/disallow process to be swapped out
/USER=                  set the username (account) the server executes under
/VERSION                simply display the server version
/[NO]ZERO               accounting zeroed on startup (default is non-zeroed)


VERSION HISTORY
---------------
See VERSION.H
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <iodef.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <libclidef.h>
#include <prvdef.h>
#include <prcdef.h>
#include <psldef.h>
#include <secdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>

/* application-related header files */
#include "wasd.h"
#include "copyright.h"

#define WASD_MODULE "HTTPD"

/******************/
/* global storage */
/******************/

/*
   These are the required privileges of the executing HTTP server.
   The server ACCOUNT should only have TMPMBX and NETMBX (just for
   extra security, policy ... keep privileged accounts to a minimum).
   The image should be installed with SYSPRV, PRMMBX, PSWAPM and SYSNAM.
   Script processes are created only with the process's current privileges,
   which are always maintained at TMPMBX and NETMBX.  If additional
   privileges are required for any particular purpose (e.g. binding to
   a privileged IP port) then they are enabled, the action performed,
   and then they are disabled again immediately.
*/
unsigned long  AltPriMask [2] = { PRV$M_ALTPRI, 0 },
               AveJoePrvMask [2] = { PRV$M_NETMBX | PRV$M_TMPMBX, 0 },
               DetachMask [2] = { PRV$M_DETACH, 0 },
#ifdef __ALPHA
               GblSecPrvMask [2] = { PRV$M_SYSPRV | PRV$M_SYSGBL |
                                     PRV$M_PRMGBL, 0 },
#else
               GblSecPrvMask [2] = { PRV$M_SYSPRV | PRV$M_SYSGBL |
                                     PRV$M_PRMGBL | PRV$M_SHMEM, 0 },
#endif
               PswapmMask [2] = { PRV$M_PSWAPM, 0 },
               ResetPrvMask [2] = { 0xffffffff, 0xffffffff },
               SysPrvMask [2] = { PRV$M_SYSPRV, 0 },
               WorldMask [2] = { PRV$M_WORLD, 0 };

char  Utility [] = "HTTPD";

/*
   Decided to be able to eliminate the debug statements from production
   executables completely.  This will eliminate some largely unused code
   from the images reducing the overall file size, but more importantly
   will eliminate the test and branch execution overhead each time a debug
   statement is encountered.  Do this by conditionally turning the integer
   Debug storage into a constant false value ... compiler optimization of
   an impossible-to-execute section of code does the rest!  Very little
   other code needs to be explicitly conditionally compiled.
*/
#ifdef DBUG
BOOL  Debug;
#else
#define Debug 0
#endif

BOOL  HttpdServerExecuting,
      HttpdTicking,
      MonitorEnabled,
      NaturalLanguageEnglish,
      NoSwapOut = true,
      OperateWithSysPrv;

int  ExitStatus,
     FileBufferSize = DEFAULT_FILE_BUFFER_SIZE,
     HttpdBytLmAvailable,
     HttpdJpiAstLm,
     HttpdJpiBioLm,
     HttpdJpiBytLm,
     HttpdJpiDioLm,
     HttpdJpiEnqLm,
     HttpdJpiFilLm,
     HttpdJpiPgFlQuo,
     HttpdJpiPrcLm,
     HttpdJpiTqLm,
     HttpdTickSecond,
     ProcessPriority = 4,
     ServerPort = 80,
     VmsVersion;

unsigned long  HttpdProcessId,
               HttpdUic;

unsigned long  HttpdStartBinTime [2];

char  HttpdUserName [13],
      NaturalLanguage [64],
      ProcessName [16],
      ProcessIdentName [32] = "WASD_",
      ServerGroupResName [32] = DEFAULT_SERVER_GROUP_RESNAME,
      ServerPortString [8],
      SyiHwName [32],
      SyiVersion [9];

/* zero index is used to indicate request is not in a list */
#define SUPERVISOR_LIST_MAX 10
SUPERVISOR_LIST  SupervisorListArray [SUPERVISOR_LIST_MAX+1] =
/* these numbers are staggered so they are not all scanned at any one time */
   { { {0,0,0}, 0, },
     { {0,0,0}, 1, },
     { {0,0,0}, 15, },
     { {0,0,0}, 50, },
     { {0,0,0}, 110, },
     { {0,0,0}, 290, },
     { {0,0,0}, 590, },
     { {0,0,0}, 1100, },
     { {0,0,0}, 3500, },
     { {0,0,0}, 7100, },
     { {0,0,0}, 999999999, } };

struct AnExitHandler  ExitHandler;

ACCOUNTING_STRUCT  *AccountingPtr;
BOOL  AccountingZeroOnStartup;

/* default storage in case the real global section cannot be/is not in use */
HTTPD_GBLSEC  HttpdGblSecDefault;
HTTPD_GBLSEC  *HttpdGblSecPtr;
int  HttpdGblSecPages;

/********************/
/* external storage */
/********************/

extern BOOL  AuthorizationEnabled,
             CliDetach,
             CliGblSecDelete,
             CliMonitorDisabled,
             CliMonitorEnabled,
             CliOdsExtendedEnabled,
             CliOdsExtendedDisabled,
             CliVersion,
             ControlDoAllHttpd,
             LoggingEnabled,
             OdsExtended,
             ProtocolHttpsConfigured;

extern int  CliServerPort,
            ConnectCountConcurrentMax,
            HttpdGblSecVersion,
            OpcomMessages,
            RequestHistoryMax,
            ServerPort,
            WatchEnabled;

extern char  BuildInfo[],
             CliLogFileName[],
             CliParameter[],
             CliProcessIdentName[],
             CliProxyMaint[],
             CliUserName[],
             ControlBuffer[],
             ErrorSanityCheck[],
             HttpdIpPackage[],
             HttpdSesola[],
             LoggingFileName[],
             Services[],
             ServerHostPort[],
             SoftwareID[],
             TcpIpAgentInfo[],
             TimeGmtString[];

extern CONFIG_STRUCT  Config;
extern LIST_HEAD  RequestList;
extern MSG_DB  MsgDb;
extern PROXY_ACCOUNTING_STRUCT  *ProxyAccountingPtr;
extern SERVICE_LOAD  ServiceLoad;

/* for initial debugging only */
extern unsigned long  VmCacheVmZoneId,
                      VmGeneralVmZoneId,
                      VmRequestVmZoneId;

/*****************************************************************************/
/*
*/

int main ()

{
   static unsigned short  SyiHwNameLength;
   static unsigned long  JpiGrp,
                         SyiMaxSysGroup;

   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *short_ret_len;
   }
   JpiItems [] =
   {
     { sizeof(JpiGrp), JPI$_GRP, &JpiGrp, 0 },
     { sizeof(HttpdProcessId), JPI$_PID, &HttpdProcessId, 0 },
     { sizeof(HttpdUic), JPI$_UIC, &HttpdUic, 0 },
     { sizeof(HttpdUserName)-1, JPI$_USERNAME, &HttpdUserName, 0 },
     { sizeof(HttpdJpiAstLm), JPI$_ASTLM, &HttpdJpiAstLm, 0 },
     { sizeof(HttpdJpiBioLm), JPI$_BIOLM, &HttpdJpiBioLm, 0 },
     { sizeof(HttpdJpiBytLm), JPI$_BYTLM, &HttpdJpiBytLm, 0 },
     { sizeof(HttpdJpiDioLm), JPI$_DIOLM, &HttpdJpiDioLm, 0 },
     { sizeof(HttpdJpiEnqLm), JPI$_ENQLM, &HttpdJpiEnqLm, 0 },
     { sizeof(HttpdJpiFilLm), JPI$_FILLM, &HttpdJpiFilLm, 0 },
     { sizeof(HttpdJpiPgFlQuo), JPI$_PGFLQUOTA, &HttpdJpiPgFlQuo, 0 },
     { sizeof(HttpdJpiPrcLm), JPI$_PRCLM, &HttpdJpiPrcLm, 0 },
     { sizeof(HttpdJpiTqLm), JPI$_TQLM, &HttpdJpiTqLm, 0 },
     { 0,0,0,0 }
   },
   SyiItem [] =
   {
     { sizeof(SyiMaxSysGroup), SYI$_MAXSYSGROUP, &SyiMaxSysGroup, 0 },
     { sizeof(SyiVersion)-1, SYI$_VERSION, &SyiVersion, 0 },
     { sizeof(SyiHwName)-1, SYI$_HW_NAME, &SyiHwName, &SyiHwNameLength },
     { 0,0,0,0 }
   };

   int  status;
   unsigned short  Length;
   char  *cptr, *sptr, *zptr;
   char  TimeScratch [32];
   $DESCRIPTOR (NaturalLanguageDsc, NaturalLanguage);
   $DESCRIPTOR (ProcessNameDsc, ProcessName);

   /*********/
   /* begin */
   /*********/

   /* if no global section (yet) gotta have the storage somewhere! */
   HttpdGblSecPtr = &HttpdGblSecDefault;
   HttpdGblSecPages = sizeof(HTTPD_GBLSEC) / 512;
   if (HttpdGblSecPages & 0x1ff) HttpdGblSecPages++;
   AccountingPtr = &HttpdGblSecPtr->Accounting;
   ProxyAccountingPtr = &HttpdGblSecPtr->ProxyAccounting;

   if (VMSnok (status = ParseCommandLine ()))
      exit (status);

#ifdef DBUG
   if (!Debug)
      if ((char*)getenv ("HTTPD$DBUG") != NULL)
         Debug = true;
#endif

   if (ControlBuffer[0])
   {
      /*******************************************/
      /* being used to control the HTTPd process */
      /*******************************************/

      exit (ControlCommand (ControlBuffer));
   }

   if (CliDetach)
   {
      /************************************/
      /* create a detached server process */
      /************************************/

      exit (HttpdDetachServerProcess ());
   }
   
   VersionInfo ();

   /* output the GNU GENERAL PUBLIC LICENSE message */
   WriteFaoStdout ("%!AZ-I-SOFTWAREID, !AZ\n!AZ",
                   Utility, SoftwareID, CopyRightMessageBrief);

   if (Debug)
      fprintf (stdout,
"AccountingStruct: %d\n\
CacheStruct: %d\n\
DclTaskStruct: %d\n\
DclScriptNameCacheStruct: %d\n\
DirTaskStruct: %d\n\
DECnetConnectStruct: %d\n\
DECnetTaskStruct: %d\n\
HttpdGblSecStruct: %d\n\
MonRequestStruct: %d\n\
OdsStruct: %d\n\
ProxyTaskStruct: %d\n\
ProxyAccountingStruct: %d\n\
PutTaskStruct: %d\n\
RequestStruct: %d\n\
ServiceStruct: %d\n\
SsiStruct: %d\n\
UpdTaskStruct: %d\n",
               sizeof(ACCOUNTING_STRUCT),
               sizeof(struct CacheStruct),
               sizeof(struct DclTaskStruct),
               sizeof(struct DclScriptNameCacheStruct),
               sizeof(struct DirTaskStruct),
               sizeof(struct DECnetConnectStruct),
               sizeof(struct DECnetTaskStruct),
               sizeof(struct HttpdGblSecStruct),
               sizeof(struct MonRequestStruct),
               sizeof(struct OdsStruct),
               sizeof(struct ProxyTaskStruct),
               sizeof(PROXY_ACCOUNTING_STRUCT),
               sizeof(struct PutTaskStruct),
               sizeof(struct RequestStruct),
               sizeof(struct ServiceStruct),
               sizeof(struct SsiTaskStruct),
               sizeof(struct UpdTaskStruct));

   if (CliVersion) exit (SS$_NORMAL);

   /**********************/
   /* proxy maintainance */
   /**********************/

   if (CliProxyMaint[0])
   {
      ProxyMaintCli ();
      exit (SS$_NORMAL);
   }

   /* quick hack to delete the permanent global section used for monitoring */
   if (CliGblSecDelete) HttpdGblSecInit (true);

   /****************/
   /* HTTPd server */
   /****************/

   /* timestamp the server startup */
   sys$gettim (&HttpdStartBinTime);
   WriteFaoStdout ("%!AZ-I-STARTUP, !20%D\n", Utility, 0);

   /* display information about any CLI WATCH facility settings */
   WatchCliSettings ();

   /* set up and declare the exit handler */
   ExitHandler.HandlerAddress = &HttpdExit;
   ExitHandler.ArgCount = 1;
   ExitHandler.ExitStatusPtr = &ExitStatus;
   if (VMSnok (status = sys$dclexh (&ExitHandler)))
      ErrorExitVmsStatus (status, "sys$dclexh()", FI_LI);

   HttpdServerExecuting = true;
   HttpdOnControlY (false);

   /* initialize general-use virtual memory management */
   VmGeneralInit ();
   /** if (Debug) VmDebug (VmGeneralVmZoneId); **/

   /* get some details of the account */
   if (VMSnok (status = sys$getjpiw (0, 0, 0, &JpiItems, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$getjpiw()", FI_LI);
   HttpdUserName[sizeof(HttpdUserName)-1] = '\0';
   for (cptr = HttpdUserName; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   if (Debug)
      fprintf (stdout,
         "Uic: %08.08X UserName |%s| ProcessId: %08.08X Grp: %d\n",
          HttpdUic, HttpdUserName, HttpdProcessId, JpiGrp);

   /* get system information */
   if (VMSnok (status = sys$getsyi (0, 0, 0, &SyiItem, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$getsyi()", FI_LI);

   SyiHwName[SyiHwNameLength] = '\0';
   SyiVersion[sizeof(SyiVersion)-1] = '\0';
   for (cptr = SyiVersion; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   VmsVersion = ((SyiVersion[1]-48) * 10) + (SyiVersion[3]-48);
   if (Debug)
      fprintf (stdout, "HwName |%s| SyiVersion |%s| VmsVersion: %d\n",
               SyiHwName, SyiVersion, VmsVersion);

   WriteFaoStdout ("%!AZ-I-SYSTEM, !AZ VMS !AZ\n",
                   Utility, SyiHwName, SyiVersion);

   if (JpiGrp <= SyiMaxSysGroup)
      WriteFaoStdout (
"%!AZ-W-SYSPRV, operating with implicit SYSPRV (UIC group !UL)\n",
         Utility, JpiGrp);

   /* get the TCP/IP agent information */
   NetTcpIpAgentInfo ();
   WriteFaoStdout ("%!AZ-I-TCPIP, !AZ\n", Utility, TcpIpAgentInfo);

   /* set the flag and report indicating whether ODS-5 is supported */
   OdsSetExtended ();

   /* make sure the process' privileges are those of a mere mortal */
   if (VMSnok (status = sys$setprv (0, &ResetPrvMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   if (VMSnok (status = sys$setprv (1, &AveJoePrvMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);

   if (OperateWithSysPrv && OPERATE_WITH_SYSPRV)
   {
      /* looks like we're been asked to use Superman's Xray vision! */
      if (VMSnok (status = sys$setprv (1, &SysPrvMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
      WriteFaoStdout ("%!AZ-I-SYSPRV, operating with explicit SYSPRV\n",
                      Utility);
   }
   else
   if (OperateWithSysPrv)
   {
      OperateWithSysPrv = false;
      WriteFaoStdout ("%!AZ-W-SYSPRV, operation is NOT a compiled option\n",
                      Utility);
   }

   /* set (a sensible) process priority */
   if (ProcessPriority > 6) ProcessPriority = 6;
   if (VMSnok (status = sys$setprv (1, &AltPriMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   if (VMSnok (status = sys$setpri (0, 0, ProcessPriority, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setpri()", FI_LI);
   if (VMSnok (status = sys$setprv (0, &AltPriMask, 0, 0)))
      ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);

   /* anywhere near before starting logging! */
   NetGetServerHostName ();

   /* read the server configuration file */
   Configure (&Config);

   if (CliServerPort)
      ServerPort = CliServerPort;
   else
   if (Config.cfServer.DefaultPort)
      ServerPort = Config.cfServer.DefaultPort;
   if (ServerPort < 1 || ServerPort > 65535)
      ErrorExitVmsStatus (0, "IP port", FI_LI);
   sprintf (ServerPortString, "%d", ServerPort);

   WriteFao (HttpdGblSecPtr->StatusMessage,
             sizeof(HttpdGblSecPtr->StatusMessage), NULL,
             "%!AZ-I-STARTUP, !20%D", Utility, 0);

   /* initialize Secure Sockets Layer, if available and if required */
   SesolaInit ();

   /* get the configured services */
   ServiceConfigLoad(&ServiceLoad);

   /* set/reset global section used by CLI control and the HTTPMON utility */
   HttpdGblSecInit (false);

   /* update (reset) the request information */
   if (MonitorEnabled) RequestGblSecUpdate (NULL);

   /* now that we have the accounting data (in the global section) */
   if (AccountingZeroOnStartup) ZeroAccounting ();

   /* set the process name */
   status = WriteFao (ProcessName, sizeof(ProcessName), &Length,
                      "HTTPd:!UL", ServerPort);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorExitVmsStatus (status, "WriteFao()", FI_LI);
   if (Debug) fprintf (stdout, "ProcessName |%s|\n", ProcessName);
   ProcessNameDsc.dsc$w_length = Length;
   if (VMSnok (status = sys$setprn (&ProcessNameDsc)))
      ErrorExitVmsStatus (status, "sys$setprn()", FI_LI);

   /* not the process name has been set OPCOM can be used */
   if (OpcomMessages & OPCOM_HTTPD)
      WriteFaoOpcom ("%!AZ-I-STARTUP, !AZ", Utility, SoftwareID);

   /* set up a rights identifier name (currently for detached scripting) */
   zptr = (sptr = ProcessIdentName) + sizeof(ProcessIdentName)-1;
   if (CliProcessIdentName[0])
      for (cptr = CliProcessIdentName; *cptr && sptr < zptr; *sptr++ = *cptr++);
   else
   {
      /* server-set rights identifiers begin with the following */
      for (cptr = "WASD_"; *cptr; *sptr++ = *cptr++);
      for (cptr = ProcessName; *cptr && sptr < zptr; cptr++)
         if (isalnum(*cptr)) *sptr++ = toupper(*cptr); else *sptr++ = '_';
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "ProcessIdentName |%s|\n", ProcessIdentName);

   /* ensure the GMT time/logical is available */
   if (VMSnok (status = TimeSetGMT ()))
      ErrorExitVmsStatus (status, "TimeSetGMT()", FI_LI);
   fprintf (stdout, "%%%s-I-GMT, %s\n", Utility, TimeGmtString);

   if (CliMonitorDisabled)
      MonitorEnabled = false;
   else
   if (Config.cfMisc.MonitorEnabled || CliMonitorEnabled)
      MonitorEnabled = true;
   else
      MonitorEnabled = false;

   /* initialize message database */
   MsgLoad (&MsgDb);

   /* check some contents of that message database */
   ErrorCheckReportFormats ();

   /* authentication/authorization configuration */
   AuthConfigInit ();

   /* load the rule mapping database */
   MapUrl_Load ();

   /* initialize proxy processing */
   ProxyInit ();

   /* initialize DCL processing */
   DclInit ();

   if (Config.cfServer.BusyLimit)
      ConnectCountConcurrentMax = Config.cfServer.BusyLimit;

   /* initialize request virtual memory management */
   VmRequestInit (ConnectCountConcurrentMax); 
   /** if (Debug) VmDebug (VmRequestVmZoneId); **/

   /* initialize file cache  */
   CacheInit (true);
   /** if (Debug) VmDebug (VmCacheVmZoneId); **/
   /** if (Debug) VmDebug (0); **/

   /* initialize activity statistics */
   GraphActivityInit ();

   /* initialize the language environment */
   if (VMSok (status = lib$get_users_language (&NaturalLanguageDsc)))
   {
      for (cptr = NaturalLanguage; *cptr && !ISLWS(*cptr); cptr++);
      *cptr = '\0';
      if (strsame (NaturalLanguage, "ENGLISH"))
         NaturalLanguageEnglish = true;
      else
      {
         NaturalLanguageEnglish = false;
         fprintf (stdout, "%%%s-I-LANGUAGE, natural language is %s\n",
                  Utility, NaturalLanguage);
      }
   }
   else
   {
      if (status == LIB$_ENGLUSED)
      {
         NaturalLanguageEnglish = true;
         strcpy (NaturalLanguage, "ENGLISH");
         fprintf (stdout,
            "%%%s-W-LANGUAGE, natural language has defaulted to %s\n",
            Utility, NaturalLanguage);
      }
      else
         ErrorExitVmsStatus (status, "lib$get_users_language()", FI_LI);
   }

   /* set up for the HTTPD to be controlled via DLM and global section */
   ControlInit ();

   /* initialize request history mechanism (ensure it's reasonable!) */
   RequestHistoryMax = Config.cfMisc.RequestHistory;
   if (RequestHistoryMax > 999) RequestHistoryMax = 0;

   /* disable process swapping */
   if (NoSwapOut)
   {
      if (VMSnok (status = sys$setprv (1, &PswapmMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
      if (VMSnok (status = sys$setswm (1)))
         ErrorExitVmsStatus (status, "sys$setswm()", FI_LI);
      if (VMSnok (status = sys$setprv (0, &PswapmMask, 0, 0)))
         ErrorExitVmsStatus (status, "sys$setprv()", FI_LI);
   }

   /* disable AST's until we've created the listening sockets */
   sys$setast (0);

   /* need SYSPRV to bind to a well-known port */
   EnableSysPrv();

   /* create network services */
   NetCreateService ();

   /* turn off SYSPRV after binding to well-known port */
   DisableSysPrv();

   /* initialize logging after service creation (in case of per-service logs) */
   if (VMSnok (status = Logging (NULL, LOGGING_BEGIN)))
      ErrorExitVmsStatus (status, "Logging()", FI_LI);

   /* force ODS extended if the command-line requested it (testing only) */
   if (CliOdsExtendedEnabled) OdsExtended = true;
   if (CliOdsExtendedDisabled) OdsExtended = false;

   /* get the available BYTLM quota after all server sockets created */
   HttpdBytLmAvailable = GetJpiBytLm();
   if (Debug) fprintf (stdout, "BytLmAvailable: %d\n", HttpdBytLmAvailable);

   /*****************************/
   /* begin to process requests */
   /*****************************/

   WriteFaoStdout ("%!AZ-I-BEGIN, !20%D, accepting requests\n", Utility, 0);

   /* cancel any startup messages provided for the monitor */
   HttpdGblSecPtr->StatusMessage[0] = '\0';

   /* reenable user-mode ASTs to allow use of our services */
   sys$setast (1);

   sys$hiber ();

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
Create a detached server process.  If the /USER= qualifier was used the process
will  have changed it's persona and the process created here will be created
under that user name.
*/

HttpdDetachServerProcess ()

{
   static unsigned long  CrePrcFlags = PRC$M_DETACH;
   static $DESCRIPTOR (LoginOutDsc, "SYS$SYSTEM:LOGINOUT.EXE");
   static $DESCRIPTOR (SysCommandDsc, "HT_ROOT:[LOCAL]STARTUP_SERVER.COM");
   static $DESCRIPTOR (SysOutputFaoDsc, "HT_ROOT:[LOG.SERVER]!AZ.LOG");
   static $DESCRIPTOR (SysOutputDsc, "");

   static unsigned short  Length;
   static char  SyiNodeName [16];
   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *short_ret_len;
   }
   SyiItem [] =
   {
     { sizeof(SyiNodeName)-1, SYI$_NODENAME, &SyiNodeName, &Length },
     { 0,0,0,0 }
   };

   int  status;
   unsigned int  ProcessPid;
   char  LogFileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpDetachServerProcess()\n");

   if (CliUserName[0])
      if (VMSnok (status = PersonaAssume (CliUserName)))
         exit (status);

   if (CliParameter[0])
   {
      SysCommandDsc.dsc$a_pointer = CliParameter;
      SysCommandDsc.dsc$w_length = strlen(CliParameter);
   }

   if (LoggingFileName[0])
   {
      SysOutputDsc.dsc$a_pointer = LoggingFileName;
      SysOutputDsc.dsc$w_length = strlen(LoggingFileName);
   }
   else
   {
      if (VMSnok (status = sys$getsyi (0, 0, 0, &SyiItem, 0, 0, 0)))
         ErrorExitVmsStatus (status, "sys$getsyi()", FI_LI);
      SyiNodeName[Length] = '\0';
      WriteFao (LogFileName, sizeof(LogFileName), &Length,
                "HT_ROOT:[LOG.SERVER]!AZ.LOG", SyiNodeName);
      SysOutputDsc.dsc$a_pointer = LogFileName;
      SysOutputDsc.dsc$w_length = Length;
   }

   status = sys$creprc (&ProcessPid,
                        &LoginOutDsc,
                        &SysCommandDsc,
                        &SysOutputDsc,
                        0, 0, 0, 0, ProcessPriority, 0, 0,
                        CrePrcFlags,
                        0, 0);
   if (VMSnok (status)) exit (status);

   WriteFaoStdout (
      "%!AZ-S-PROC_ID, identification of created process is !8XL\n",
      Utility, ProcessPid);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Exit the HTTP server, via this declared exit handler.
*/

HttpdExit (unsigned long *ExitStatusPtr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdExit() %%X%08.08X\n", *ExitStatusPtr);

   /* run down any script processes associated with script processing */
   DclExit ();

   if (LoggingEnabled)
   {
      /* provide a server exit entry */
      Logging (NULL, LOGGING_END);
   }


   AccountingPtr->LastExitStatus = *ExitStatusPtr;

   if (VMSnok (*ExitStatusPtr))
   {
      /* report to process log */
      WriteFaoStdout ("%!AZ-F-EXIT, %X!8XL\n",
                      Utility, *ExitStatusPtr);
      WriteFao (HttpdGblSecPtr->StatusMessage,
                sizeof(HttpdGblSecPtr->StatusMessage), NULL,
                "%!AZ-F-EXIT, !20%D, %X!8XL\r\n-!%M",
                Utility, 0, *ExitStatusPtr, *ExitStatusPtr);
   }
   else
      WriteFaoStdout ("%!AZ-I-EXIT, %X!8XL\n-!%M\n",
                      Utility, *ExitStatusPtr, *ExitStatusPtr);

   if (OpcomMessages)
   {
      /* an error exit is always reported via OPCOM */
      if (VMSnok (*ExitStatusPtr))
         WriteFaoOpcom ("%!AZ-F-EXIT, %X!8XL\r\n-!%M",
                        Utility, *ExitStatusPtr, *ExitStatusPtr);
      else
      if (OpcomMessages & OPCOM_HTTPD)
         WriteFaoOpcom ("%!AZ-I-EXIT, %X!8XL\r\n-!%M",
                        Utility, *ExitStatusPtr, *ExitStatusPtr);
   }
}

/*****************************************************************************/
/*
*/

int HttpdOnControlY (BOOL ControlY)

{
   static BOOL  Disabled = false;
   static unsigned long  Mask = LIB$M_CLI_CTRLY,
                         OldMask;
   static unsigned short  TTChannel = 0;

   int  status;
   $DESCRIPTOR (TTDsc, "TT:");

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdOnControlY() %d\n", ControlY);

   if (ControlY) exit (0x00000610);

   if (!TTChannel)
      if (VMSnok (status = sys$assign (&TTDsc, &TTChannel, 0, 0, 0)))
         return (status);

   if (VMSnok (status =
       sys$qiow (0, TTChannel, IO$_SETMODE | IO$M_CTRLYAST, 0, 0, 0,
                 &HttpdOnControlY, true, PSL$C_USER, 0, 0, 0)))
      return (status);

   if (VMSnok (status =
       sys$qiow (0, TTChannel, IO$_SETMODE | IO$M_CTRLCAST, 0, 0, 0,
                 &HttpdOnControlY, true, PSL$C_USER, 0, 0, 0)))
      return (status);

   if (!Disabled)
   {
      Disabled = true;
      return (lib$disable_ctrl (&Mask, &OldMask));
   }
   else
      return (status);
}

/*****************************************************************************/
/*
As the code gets more complex it's becoming increasingly possible a coding or
design error will leave privileges turned on somewhere. To help detect any such
problem occasionally (when there are no more connections to process) ensure
everthing's off that's supposed to be off.
*/

#ifdef __DECC
#pragma inline(HttpdCheckPriv)
#endif

HttpdCheckPriv ()

{
   static long  Pid = 0;
   static unsigned long  JpiCurPriv [2];
   static struct
   {
      unsigned short  buf_len;
      unsigned short  item;
      unsigned char   *buf_addr;
      unsigned short  *short_ret_len;
   }
      JpiItems [] =
   {
      { sizeof(JpiCurPriv), JPI$_CURPRIV, &JpiCurPriv, 0 },
      {0,0,0,0}
   };

   int  status;
   char  Buffer [128];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdCheckPriv()\n");

   if (VMSnok (status = sys$getjpiw (0, &Pid, 0, &JpiItems, 0, 0, 0)))
      ErrorExitVmsStatus (status, "sys$getjpiw()", FI_LI);

   if (Debug)
      fprintf (stdout, "%08.08X%08.08X\n", JpiCurPriv[1], JpiCurPriv[0]);

   JpiCurPriv[0] &= ~AveJoePrvMask[0];
   JpiCurPriv[1] &= ~AveJoePrvMask[1];

   /* if only NETMBX and TMPMBX enabled then ok */
   if (!(JpiCurPriv[0] || JpiCurPriv[1])) return;

#if OPERATE_WITH_SYSPRV
   /* if operating with SYSPRV and it's enabled then ok */
   if (OperateWithSysPrv && (JpiCurPriv[0] == SysPrvMask[0])) return;
#endif

   /* hmmmm, shouldn't have extended privileges!! */
   WriteFao (Buffer, sizeof(Buffer), 0,
             "privilege sanity check: !8XL!8XL",
             ~JpiCurPriv[1], ~JpiCurPriv[0]);
   ErrorExitVmsStatus (SS$_BUGCHECK, Buffer, FI_LI);
}

/*****************************************************************************/
/*
When in use this function calls itself (via a timer AST) every second.  It
updates the 'HttpdTickSecond' global storage by using system time as a baseline
(which works around any latency or cluster transition issues, etc.)  This
counts upwards across year boundaries.  It then calls the request supervisor,
and then the DCL task supervisor.  If either of these has requests or tasks
still to supervise they do it an return a true which indicates the ticker
should continue to tick.  Part of the duty of this function is to update the
global storage 'HttpdTickSecond' which indicates the timeline progress of the
server as well as is used by various sections to set expiry points (marked by
these global "seconds").  Ticking is initiated by RequestBegin() calling this
function with a (true) parameter upon the first request processed in an
otherwise quiescent server ('HttpdTicking' is false).
*/

HttpdTick (BOOL BeginTicking)

{
   static unsigned long  LibSecondOfYear = LIB$K_SECOND_OF_YEAR;
   static unsigned long  OneSecondDelta [2] = { -10000000, -1 };
   static int  AdjustSecondOfYear = 0,
               PrevMinuteOfYear = 0,
               PrevSecondOfYear = 0;

   int  status;
   unsigned long  SecondOfYear,
                  MinuteOfYear;
   unsigned long  BinTime [2];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "HttpdTick() %d %d\n", BeginTicking, HttpdTickSecond);

   /* get the absolute second count (work-around any timer AST latency) */
   sys$gettim (&BinTime);
   lib$cvt_from_internal_time (&LibSecondOfYear, &SecondOfYear, &BinTime);
   if (SecondOfYear < PrevSecondOfYear)
   {
      /* happy new year! (the Web never sleeps ;^) */
      AdjustSecondOfYear += 31622400;
   }
   PrevSecondOfYear = SecondOfYear;
   MinuteOfYear = SecondOfYear / 60;
   HttpdTickSecond = AdjustSecondOfYear + SecondOfYear;
   if (Debug)
      fprintf (stdout, "%d %d %d %d %d\n",
               AdjustSecondOfYear, SecondOfYear, MinuteOfYear,
               PrevSecondOfYear, HttpdTickSecond);

/**
   if (WatchEnabled & WATCH_TIMER && WatchEnabled & WATCH_DETAIL)
      WatchThis (NULL, FI_LI, WATCH_TIMER+WATCH_DETAIL,
                 "TICK !UL", HttpdTickSecond);
**/

   if (BeginTicking)
      HttpdTicking = true;
   else
   {
      HttpdTicking = false;
      if (HttpdSupervisor()) HttpdTicking = true;
      if (DclSupervisor (-1)) HttpdTicking = true;
      if (DECnetSupervisor (-1)) HttpdTicking = true;
      if (!HttpdTicking || MinuteOfYear != PrevMinuteOfYear)
         ThrottleMonitorReset ();
   }

   if (MinuteOfYear != PrevMinuteOfYear) PrevMinuteOfYear = MinuteOfYear;

   if (HttpdTicking)
      if (VMSnok (status =
          sys$setimr (0, &OneSecondDelta, &HttpdTick, false, 0)))
         ErrorExitVmsStatus (status, "sys$setimr()", FI_LI);
}

/*****************************************************************************/
/*
Set the timer counters for the request according to the function code. 
HttpdSupervisor() will monitor these counters.  The period is set by adding the
period in seconds to the current second count to derive a value in the furture. 
When this is reached or exceeded the timer has expired.   The 'DurationSeconds'
is an optional parameter.  If set to zero the standard timer values are used. 
If non-zero the value is used.  This allows CGI callouts and scripting control
to set these with specific values.
*/

HttpdTimerSet
(
REQUEST_STRUCT *rqptr,
int Function,
int DurationSeconds
)
{
   static BOOL  Initialize = true;
   static unsigned long  InputSeconds,
                         KeepAliveSeconds,
                         NoProgressPeriod,
                         OutputSeconds;

   int  idx, status,
        TimerSeconds;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdTimerSet() %d\n", Function);

   if (Initialize)
   {
      Initialize = false;
      if (Config.cfTimeout.Input)
         InputSeconds = Config.cfTimeout.Input;
      else
         InputSeconds = DEFAULT_TIMEOUT_INPUT_MINUTES * 60;
      if (Config.cfTimeout.Output)
         OutputSeconds = Config.cfTimeout.Output;
      else
         OutputSeconds = DEFAULT_TIMEOUT_OUTPUT_MINUTES * 60;
      if (Config.cfTimeout.NoProgress)
         NoProgressPeriod = Config.cfTimeout.NoProgress;
      else
         NoProgressPeriod = DEFAULT_TIMEOUT_NOPROGRESS_MINUTES * 60;
      if (Config.cfTimeout.KeepAlive)
         KeepAliveSeconds = Config.cfTimeout.KeepAlive;
      else
         KeepAliveSeconds = DEFAULT_TIMEOUT_KEEPALIVE_SECONDS;
      /* ensure small periods are at least that duration */
      if (KeepAliveSeconds && KeepAliveSeconds < 10) KeepAliveSeconds++;
   }

   switch (Function)
   {
      case TIMER_INPUT :

         rqptr->rqTmr.KeepAliveSecond =
            rqptr->rqTmr.NoProgressBytesRx =
            rqptr->rqTmr.NoProgressBytesTx =
            rqptr->rqTmr.NoProgressSecond =
            rqptr->rqTmr.NoProgressPeriod =
            rqptr->rqTmr.OutputSecond =
            rqptr->rqTmr.ThrottleCount = 0;

         if (DurationSeconds)
            TimerSeconds = DurationSeconds;
         else
            TimerSeconds = InputSeconds;

         if (rqptr->WatchItem &&
            (WatchEnabled & WATCH_TIMER))
            WatchThis (rqptr, FI_LI, WATCH_TIMER,
                       "INPUT !UL seconds", TimerSeconds);

         rqptr->rqTmr.InputSecond = HttpdTickSecond + TimerSeconds + 1;

         break;

      case TIMER_KEEPALIVE :

         rqptr->rqTmr.InputSecond =
            rqptr->rqTmr.OutputSecond =
            rqptr->rqTmr.NoProgressBytesRx =
            rqptr->rqTmr.NoProgressBytesTx =
            rqptr->rqTmr.NoProgressSecond =
            rqptr->rqTmr.NoProgressPeriod =
            rqptr->rqTmr.ThrottleCount = 0;

         TimerSeconds = KeepAliveSeconds;

         if (rqptr->WatchItem &&
            (WatchEnabled & WATCH_TIMER))
            WatchThis (rqptr, FI_LI, WATCH_TIMER,
                       "KEEP-ALIVE !UL seconds", TimerSeconds);

         rqptr->rqTmr.KeepAliveSecond = HttpdTickSecond + KeepAliveSeconds + 1;

         break;

      case TIMER_NOPROGRESS :
      case TIMER_OUTPUT :

         rqptr->rqTmr.InputSecond =
            rqptr->rqTmr.KeepAliveSecond =
            rqptr->rqTmr.ThrottleCount = 0;

         /* establish the baseline no-progress indicators */
         rqptr->rqTmr.NoProgressBytesRx = rqptr->BytesRx;
         rqptr->rqTmr.NoProgressBytesTx = rqptr->BytesTx;

         if (Function == TIMER_NOPROGRESS && DurationSeconds)
            rqptr->rqTmr.NoProgressPeriod = DurationSeconds;
         else
         if (rqptr->rqPathSet.TimeoutNoProgress)
            rqptr->rqTmr.NoProgressPeriod = rqptr->rqPathSet.TimeoutNoProgress;
         else
            rqptr->rqTmr.NoProgressPeriod = NoProgressPeriod;

         if (Function == TIMER_OUTPUT && DurationSeconds)
            TimerSeconds = DurationSeconds;
         else
         if (rqptr->rqPathSet.TimeoutOutput)
            TimerSeconds = rqptr->rqPathSet.TimeoutOutput;
         else
            TimerSeconds = OutputSeconds;

         if (rqptr->WatchItem &&
            (WatchEnabled & WATCH_TIMER))
            WatchThis (rqptr, FI_LI, WATCH_TIMER,
                       "OUTPUT/NOPROGRESS !UL/!UL seconds",
                        TimerSeconds, rqptr->rqTmr.NoProgressPeriod);

         rqptr->rqTmr.NoProgressSecond =
            HttpdTickSecond + rqptr->rqTmr.NoProgressPeriod + 1;

         rqptr->rqTmr.OutputSecond = HttpdTickSecond + TimerSeconds + 1;

         /* use the smaller of the two values to establish the list */
         if (rqptr->rqTmr.NoProgressPeriod < TimerSeconds)
            TimerSeconds = rqptr->rqTmr.NoProgressPeriod;

         break;

      case TIMER_THROTTLE :

         rqptr->rqTmr.InputSecond =
            rqptr->rqTmr.KeepAliveSecond =
            rqptr->rqTmr.NoProgressBytesRx =
            rqptr->rqTmr.NoProgressBytesTx =
            rqptr->rqTmr.NoProgressSecond =
            rqptr->rqTmr.NoProgressPeriod =
            rqptr->rqTmr.OutputSecond = 0;

         if (DurationSeconds)
            TimerSeconds = DurationSeconds;
         else
         /* timeout-busy only becomes affective after any timeout-queue */
         if (rqptr->rqPathSet.ThrottleTimeoutQueue)
            TimerSeconds = rqptr->rqPathSet.ThrottleTimeoutQueue;
         else
         if (rqptr->rqPathSet.ThrottleTimeoutBusy)
            TimerSeconds = rqptr->rqPathSet.ThrottleTimeoutBusy;
         else
            TimerSeconds = OutputSeconds;

         if (rqptr->WatchItem &&
            (WatchEnabled & WATCH_TIMER))
            WatchThis (rqptr, FI_LI, WATCH_TIMER,
                       "THROTTLE !UL seconds", TimerSeconds);

         rqptr->rqTmr.ThrottleCount = HttpdTickSecond + TimerSeconds + 1;

         break;

      default :
         ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);
   }

   /* add/move to new list */
   HttpdSupervisorList (rqptr, TimerSeconds);

   rqptr->rqTmr.TerminatedCount = 0;
}

/*****************************************************************************/
/*
Whenever at least one request is being processed this function is called every
second by HttpdTick() to supervise the requests' progress.  An array of lists
is maintained so that the duration of the scan is somewhat limited each second,
peaks occuring only when non-one-second lists need scanning.  Timer duration
counts are set by adding the timer period in seconds to the current second
count to get a second count in the future.  When this is reached  or exceeded
the timer has expired.  Time queue entries are moved from list to list as the
remaining period in seconds changes, ensuring that the entry is checked at
appropriate intervals.  Return true to indicate the HTTPd should continue to
tick.
*/

boolean HttpdSupervisor ()

{
   int  idx, status,
        ChunkSeconds,
        DeltaSecondCount,
        TimerSecond;
   char  *TimerTypePtr;
   REQUEST_STRUCT  *rqeptr;
   LIST_ENTRY  *leptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdSupervisor()\n");

   if (RequestList.HeadPtr == NULL) return (false);

/**
   if (WatchEnabled & WATCH_TIMER && WatchEnabled & WATCH_DETAIL)
   {
      WatchThis (NULL, FI_LI, WATCH_TIMER+WATCH_DETAIL,
                 "SUPERVISOR !UL", HttpdTickSecond);
      WatchDataFormatted ("SECONDS COUNT SCAN\n");
      for (idx = 1; idx <= SUPERVISOR_LIST_MAX; idx++)
      {
         ChunkSeconds = SupervisorListArray[idx].ChunkSeconds;
         WatchDataFormatted ("!7UL !5UL !AZ\n",
            SupervisorListArray[idx].ChunkSeconds,
            LIST_GET_COUNT (&SupervisorListArray[idx].RequestList),
            (idx == 1 || !(HttpdTickSecond % ChunkSeconds)) ? " YES" : "  no");
      }
   }
**/

   /***************************/
   /* scan supervisor list(s) */
   /***************************/

   for (idx = 1; idx <= SUPERVISOR_LIST_MAX; idx++)
   {
      ChunkSeconds = SupervisorListArray[idx].ChunkSeconds;
      if (!(idx == 1 || !(HttpdTickSecond % ChunkSeconds))) continue;
      DeltaSecondCount = HttpdTickSecond + ChunkSeconds;

      /* process the current request list entries */
      leptr = SupervisorListArray[idx].RequestList.HeadPtr;
      while (leptr != NULL)
      {
         rqeptr = (REQUEST_STRUCT*)leptr->DataPtr;

         if (Debug)
            fprintf (stdout,
"%d <- %d -> %d (%d) in: %d o:%d np:%d ka:%d term:%d\n",
               leptr->PrevPtr, leptr, leptr->NextPtr, rqeptr,
               rqeptr->rqTmr.InputSecond, rqeptr->rqTmr.OutputSecond,
               rqeptr->rqTmr.NoProgressSecond, rqeptr->rqTmr.KeepAliveSecond,
               rqeptr->rqTmr.TerminatedCount);

         /* IMMEDIATELY get a pointer to the next in the list */
         leptr = leptr->NextPtr;

         /* ordered in the most common occurance of timings */
         if (rqeptr->rqTmr.OutputSecond)
         {
            /*********************/
            /* output/noprogress */
            /*********************/

            if (rqeptr->rqTmr.NoProgressBytesTx != rqeptr->BytesTx ||
                rqeptr->rqTmr.NoProgressBytesRx != rqeptr->BytesRx)
            {
               rqeptr->rqTmr.NoProgressSecond =
                  HttpdTickSecond + rqeptr->rqTmr.NoProgressPeriod + 1;
               rqeptr->rqTmr.NoProgressBytesRx = rqeptr->BytesRx;
               rqeptr->rqTmr.NoProgressBytesTx = rqeptr->BytesTx;
            }

            /* use the lesser of output and no-progress counts */
            if (rqeptr->rqTmr.OutputSecond <= rqeptr->rqTmr.NoProgressSecond)
            {
               if (rqeptr->rqTmr.OutputSecond > DeltaSecondCount) continue;
               TimerSecond = rqeptr->rqTmr.OutputSecond;
               TimerTypePtr = "OUTPUT";
            }
            else
            {
               if (rqeptr->rqTmr.NoProgressSecond > DeltaSecondCount) continue;
               TimerSecond = rqeptr->rqTmr.NoProgressSecond;
               TimerTypePtr = "NO-PROGRESS";
            }

            if (TimerSecond > HttpdTickSecond)
            {
               /* move to new list */
               HttpdSupervisorList (rqeptr, TimerSecond-HttpdTickSecond);
               continue;
            }
            /* fall through to 'TerminatedCount' */
         }
         else
         if (TimerSecond = rqeptr->rqTmr.InputSecond)
         {
            /*********/
            /* input */
            /*********/

            if (TimerSecond > DeltaSecondCount) continue;
            if (TimerSecond > HttpdTickSecond)
            {
               /* move to new list */
               HttpdSupervisorList (rqeptr, TimerSecond-HttpdTickSecond);
               continue;
            }
            TimerTypePtr = "INPUT";
            /* fall through to 'TerminatedCount' */
         }
         else
         if (TimerSecond = rqeptr->rqTmr.KeepAliveSecond)
         {
            /**************/
            /* keep-alive */
            /**************/

            if (TimerSecond > DeltaSecondCount) continue;
            if (TimerSecond > HttpdTickSecond)
            {
               /* move to new list */
               HttpdSupervisorList (rqeptr, TimerSecond-HttpdTickSecond);
               continue;
            }
            TimerTypePtr = "KEEP-ALIVE";
            /* fall through to 'TerminatedCount' */
         }
         else
         if (TimerSecond = rqeptr->rqTmr.ThrottleCount)
         {
            /************/
            /* throttle */
            /************/

            if (TimerSecond > DeltaSecondCount) continue;
            if (TimerSecond > HttpdTickSecond)
            {
               /* move to new list */
               HttpdSupervisorList (rqeptr, TimerSecond-HttpdTickSecond);
               continue;
            }

            if (rqeptr->WatchItem && (WatchEnabled & WATCH_TIMER))
               WatchThis (rqeptr, FI_LI, WATCH_TIMER, "THROTTLE expired");

            ThrottleTimeout (rqeptr);

            /* does not fall through to 'TerminatedCount'! */
            continue;
         }

         /*************************/
         /* shut down the request */
         /*************************/

         if (!rqeptr->rqTmr.TerminatedCount++)
         {
            if (rqeptr->WatchItem && (WatchEnabled & WATCH_TIMER))
               WatchThis (rqeptr, FI_LI, WATCH_TIMER,
                          "!AZ expired", TimerTypePtr);

            if (rqeptr->rqNet.ClientChannel) NetCloseSocket (rqeptr);
            if (rqeptr->DclTaskPtr != NULL)
            {
               /* DCL involved, do any DCL-specific cleanup */
               rqeptr->DclTaskPtr->DeleteProcess = true;
               DclTaskRunDown (rqeptr->DclTaskPtr);
            }
            else
            if (rqeptr->DECnetTaskPtr != NULL)
            {
               /* DECnet (CGI or OSU) involved */
               DECnetDisconnectTask (rqeptr->DECnetTaskPtr);
            }
            else
            if (rqeptr->ProxyTaskPtr != NULL)
            {
               if (rqeptr->ProxyTaskPtr->ProxyChannel)
                  ProxyCloseSocket (rqeptr->ProxyTaskPtr);
            }
         }
      }
   }

   return (true);
}

/*****************************************************************************/
/*
If necessary remove a request entry from it's appropriate supervisor array
list.  The search for an element within the range of timer currently still
outstanding with the request and add it to the HEAD of that list.  The head is
important because HttpdSupervisor() scans through these lists from head to tail
and if the tail is played with at the same time all hell breaks loose.
*/

HttpdSupervisorList
(
REQUEST_STRUCT *rqptr,
int TimerSeconds
)
{
   int  idx;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdSupervisorList() %d\n", TimerSeconds);

   if (rqptr->rqTmr.ListIndex)
   {
      /* remove from current list */
      ListRemove (&SupervisorListArray[rqptr->rqTmr.ListIndex].RequestList,
                  &rqptr->rqTmr.ListEntry);
      rqptr->rqTmr.ListIndex = 0;
      rqptr->rqTmr.ListEntry.DataPtr = NULL;
   }

   /* just removing from timer list */
   if (TimerSeconds == -1) return;

   /* add to new list */
   for (idx = 1; idx < SUPERVISOR_LIST_MAX; idx++)
   {
      if (TimerSeconds > SupervisorListArray[idx+1].ChunkSeconds)
         continue;
      /* add this to the head of the list (less carpet pulling that way) */
      ListAddHead (&SupervisorListArray[idx].RequestList,
                   &rqptr->rqTmr.ListEntry);
      rqptr->rqTmr.ListEntry.DataPtr = (void*)rqptr;
      rqptr->rqTmr.ListIndex = idx;
      return;
   }

   ErrorNoticed (0, "supervisor list", FI_LI);

   /* ensure it goes somewhere */
   ListAddHead (&SupervisorListArray[1].RequestList, &rqptr->rqTmr.ListEntry);
   rqptr->rqTmr.ListEntry.DataPtr = (void*)rqptr;
   rqptr->rqTmr.ListIndex = 1;
}

/*****************************************************************************/
/*
Provide a report on the server supervisor (do tell!)
*/ 

HttpdSupervisorReport
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction
)
{
   static char  ResponseFao [] =
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Supervisor Report</TITLE>\n\
</HEAD>\n\
!AZ\n\
<H2><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H3>Supervisor Report</H2>\n\
!20%W\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH ALIGN=right>List</TH><TH>Seconds</TH>\
<TH>Count</TH><TH>Scan</TH></TR>\n\
<TR></TR>\n";

   static char  ListFao [] =
"<TR><TH ALIGN=right>&nbsp;!UL&nbsp;</TH>\
<TD ALIGN=right>!7UL</TD><TD ALIGN=right>!UL</TD>\
<TD ALIGN=right>!AZ</TD></TR>\n";

   static char  EndOfPageFao [] =
"<TR></TR>\n\
<TR><TH ALIGN=right>Total&nbsp;</TH><TD></TD><TD ALIGN=right>!UL</TD></TR>\n\
<TR><TH ALIGN=right><B>Supervisor&nbsp;</TH><TD ALIGN=right>!UL</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n";

   int  idx, status,
        ChunkSeconds,
        Count,
        CountTotal;
   unsigned long  *vecptr;
   unsigned long  FaoVector [32];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdSupervisorReport()\n");

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = Config.cfServer.AdminBodyTag;
   *vecptr++ = ServerHostPort;
   *vecptr++ = &rqptr->rqTime.Vms64bit;

   status = NetWriteFaol (rqptr, ResponseFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   CountTotal = 0;
   for (idx = 1; idx <= SUPERVISOR_LIST_MAX; idx++)
   {
      ChunkSeconds = SupervisorListArray[idx].ChunkSeconds;
      Count = LIST_GET_COUNT (&SupervisorListArray[idx].RequestList);
      CountTotal += Count;

      vecptr = FaoVector;
      *vecptr++ = idx;
      *vecptr++ = ChunkSeconds;
      *vecptr++ = Count;
      if (idx == 1 || (HttpdTickSecond && !(HttpdTickSecond % ChunkSeconds)))
         *vecptr++ = "YES";
      else
         *vecptr++ = "no";

      status = NetWriteFaol (rqptr, ListFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   vecptr = FaoVector;
   *vecptr++ = CountTotal;
   *vecptr++ = HttpdTickSecond;

   status = NetWriteFaol (rqptr, EndOfPageFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Create and/or map (an existing) a permanent global section.  This global
section is used to contain the accounting structure and other data that might
be of interest to a separate utility such a HTTPDMON.  The utility merely maps
the section and continues to read values from the "common" memory.  This used
to be stored in logical names but this is seen as a more elegant and efficient
method.  The section name is based on the string "WASD_HTTPD_" plus the
"official" port number of the server process.  As this is a permanent global
section it can "permanently" store accounting data between server invocations
(in much the same way the old method of using logicals could).  Because it is
permanent though, and global section is created on a per-server basis, anyone
who "plays around" with servers on a whole range of "official" port numbers run
the risk of consuming all system sections and/or global pages.  To allow for
removing no-longer-required global sections this function contains a hack
allowing the server qualifier /GBLSEC=DELETE=port-number to delete the
corresponding global section.
*/ 

int HttpdGblSecInit (DoDelete)

{
   /* global, allocate space, system, in page file, permanent, writable */
   static int CreFlags = SEC$M_GBL | SEC$M_EXPREG | SEC$M_SYSGBL |
                         SEC$M_PAGFIL | SEC$M_PERM | SEC$M_WRT;
   static int DelFlags = SEC$M_SYSGBL;
   /* system & owner full access, group and world no access */
   static unsigned long  ProtectionMask = 0xff00;
   /* it is recommended to map into any virtual address in the region (P0) */
   static unsigned long  InAddr [2] = { 0x200, 0x200 };

   int  status,
        GblSecPages,
        PageCount,
        SetPrvStatus;
   short  ShortLength;
   unsigned long  RetAddr [2];
   char  HttpdGblSecName [32];
   $DESCRIPTOR (HttpdGblSecNameDsc, HttpdGblSecName);
   HTTPD_GBLSEC  *gsptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpdGblSecInit()\n");

   WriteFao (HttpdGblSecName, sizeof(HttpdGblSecName), &ShortLength,
             HTTPD_GBLSEC_NAME_FAO, ServerPort);
   HttpdGblSecNameDsc.dsc$w_length = ShortLength;

   if VMSnok ((SetPrvStatus = sys$setprv (1, &GblSecPrvMask, 0, 0)))
      ErrorExitVmsStatus (SetPrvStatus, "sys$setprv()", FI_LI);

   if (DoDelete)
   {
      if (WatchEnabled & WATCH_DETAIL)
         WatchThis (NULL, FI_LI, WATCH_DETAIL,
                    "MONITOR GBLSEC \"!AZ\"", HttpdGblSecName);

      /* delete the specified global section */
      status = sys$dgblsc (DelFlags, &HttpdGblSecNameDsc, 0);
      if (Debug) fprintf (stdout, "sys$dgblsc() %%X%08.08X\n", status);
      exit (status);
   }

   GblSecPages = sizeof(struct HttpdGblSecStruct) / 512;
   if (GblSecPages & 0x1ff) GblSecPages++;
   if (Debug) fprintf (stdout, "GblSecPage(lets)s: %d\n", GblSecPages);

   /* create and/or map the specified global section */
   status = sys$crmpsc (&InAddr, &RetAddr, 0, CreFlags,
                        &HttpdGblSecNameDsc, 0, 0, 0, GblSecPages, 0,
                        ProtectionMask, GblSecPages);
   if (Debug)
      fprintf (stdout, "sys$crmpsc() %%X%08.08X begin:%d end:%d\n",
               status, RetAddr[0], RetAddr[1]);

   if (status == SS$_NORMAL)
   {
      /* section already existed, check it's size - could be new version! */
      PageCount = (RetAddr[1]+1) - RetAddr[0] >> 9;
      if (Debug) fprintf (stdout, "(existing) PageCount: %d\n", PageCount);
      if (GblSecPages > PageCount)
      {
         /* hmmm, asking for more pages than previously, delete the current */
         status = sys$dgblsc (DelFlags, &HttpdGblSecNameDsc, 0);
         if (Debug) fprintf (stdout, "sys$dgblsc() %%X%08.08X\n", status);
         if (VMSnok (status))
            ErrorExitVmsStatus (status, "sys$dgblsc()", FI_LI);
         /* and try again */
         status = sys$crmpsc (&InAddr, &RetAddr, 0, CreFlags,
                              &HttpdGblSecNameDsc, 0, 0, 0, GblSecPages, 0,
                              ProtectionMask, GblSecPages);
         if (Debug)
            fprintf (stdout, "sys$crmpsc() %%X%08.08X begin:%d end:%d\n",
                     status, RetAddr[0], RetAddr[1]);
      }
   }

   if VMSnok ((SetPrvStatus = sys$setprv (0, &GblSecPrvMask, 0, 0)))
      ErrorExitVmsStatus (SetPrvStatus, "sys$setprv()", FI_LI);

   if (VMSok (status))
   {
      PageCount = (RetAddr[1]+1) - RetAddr[0] >> 9;
      HttpdGblSecPtr = gsptr = (HTTPD_GBLSEC*)RetAddr[0];
      HttpdGblSecPages = PageCount;
      AccountingPtr = &gsptr->Accounting;
      ProxyAccountingPtr = &gsptr->ProxyAccounting;

      /* if the global section has a different structure reset the storage */
      if (gsptr->GblSecVersion != HttpdGblSecVersion)
      {
         memset (gsptr, 0, sizeof(HTTPD_GBLSEC));
         ZeroAccounting ();
      }

      if (WatchEnabled & WATCH_DETAIL)
         WatchThis (NULL, FI_LI, WATCH_DETAIL,
            "MONITOR GBLSEC \"!AZ\" page(let)s:!UL %X!8XL %!-!%M",
            HttpdGblSecName, PageCount, status);
   }
   else
   {
      /* using existing, static storage */
      gsptr = HttpdGblSecPtr;
      WriteFaoStdout ("%!AZ-E-MONITOR, creating global section\n-!%M\n",
                      Utility, status);
   }

   gsptr->GblSecVersion = HttpdGblSecVersion;
   gsptr->HttpdProcessId = HttpdProcessId;
   strncpy (gsptr->HttpdVersion, HTTPD_VERSION, sizeof(gsptr->HttpdVersion)-1);
   gsptr->HttpdVersion[sizeof(gsptr->HttpdVersion)-1] = '\0';
   AccountingPtr->StartupCount++;

   return (status);
}

/*****************************************************************************/

